/*
 * Copyright 2002 Yrj Kari-Koskinen and Henri Sivonen.
 *
 * The contents of this file are subject to the Artistic License.
 * You may not use this file except in compliance with the Artistic
 * License (which you should have received along with this file).
 */

package dualroids;

import java.io.*;
import java.net.*;
import javax.swing.JOptionPane;

/**
 * The client-side interface to network. ClientModelConnector acts as
 * both the reader and the writer interface. The reader interface is run
 * in an independent thread. The methods which write to ObjectOutputStream
 * are synchronized and named as <code>writeFoo()</code>.
 *
 * @author Henri Sivonen, hsivonen@iki.fi
 * @author Yrj Kari-Koskinen, yrjo.kari-koskinen@hut.fi
 */
public final class ClientModelConnector implements Runnable {

    /* **************** *
     *                  *
     * Static variables *
     *                  *
     * **************** */

    /** The static variables for getPing() method */
    final static int PING_COUNT = 16,
        PING_SLEEP_TIME = 10;

    /** the ship indices */
    final static int
        SERVER_SHIP_INDEX = 0,
        CLIENT_SHIP_INDEX = 1;



    /* ******************** *
     *                      *
     * Non-static variables *
     *                      *
     * ******************** */
    
    /** the socket this connector uses */
    private Socket s;

    /** the ObjectOutputStream for this connector */
    private ObjectOutputStream oos;

    /** the ObjectInputStream for this connector */
    private ObjectInputStream ois;

    /** 
     * the clock difference between server and client:
     * clockDiff = serverTime - clientTime 
     */
    private long clockDiff;

    /**
     * The index of the ship owned by the client to which this connector
     * is connected. The ship indices are determined in the 
     * <code>SERVER/CLIENT_SHIP_INDEX</code> constants.
     */
    private int shipIndex;

    /** the client's arena */
    private Arena arena;

    /** the client's input handler */
    private InputHandler inputHandler;

    /** the client's <code>ShipStatePump</code> */
    private ShipStatePump pump; 

    /** the players' scores */
    private int[] scoreTable;

    /** the players' unused ships */
    private int[] shipsLeft;

    /** the players' names */
    private String[] playerNames;

    /** the space area  */
    private SpaceArea spaceArea;

    /** a boolean indicating whether the other player has already quit */
    private boolean otherPlayerHasQuit;

    /** a boolean indicating whether the game is over or not */
    private boolean gameOver;
    
    
    /**
     * This constructor is called from the factory method below. It 
     * initializes the network connection and the input/output streams.
     *
     * @param hostname the hostname of the server
     * @param local tells whether the server is local or not, the 
     * <code>shipIndex</code> field will be determined from it
     *
     * @throws IOException if the network fails to run the operations
     */
    private ClientModelConnector(String hostname, boolean local,
                                 String playerName) 
        throws IOException {

        int otherShipIndex;
        if (local) {
            this.shipIndex = SERVER_SHIP_INDEX;
            otherShipIndex = CLIENT_SHIP_INDEX;
        } else {
            this.shipIndex = CLIENT_SHIP_INDEX;
            otherShipIndex = SERVER_SHIP_INDEX;
        }

        this.s = new Socket(hostname, ServerModelConnector.SERVER_PORT);
        
        // initialize the streams
        this.oos = new ObjectOutputStream(s.getOutputStream());
        this.ois = new ObjectInputStream(s.getInputStream());
        
        // the default value (used in local connections)
        this.clockDiff = 0;

        this.playerNames = new String[2];
        this.playerNames[this.shipIndex] = playerName;
        this.playerNames[otherShipIndex] = "Player " + (otherShipIndex+1);

        this.scoreTable = new int[2];
        this.shipsLeft = new int[2];
        this.scoreTable[0] = this.scoreTable[1] = 0;
        this.shipsLeft[0] = this.shipsLeft[1] = Server.INITIAL_SHIPS;
        this.spaceArea = null;
        this.inputHandler = new InputHandler(this);
        this.pump = new ShipStatePump();
        new Thread(pump, "ShipStatePump").start();

        this.otherPlayerHasQuit = this.gameOver = false;
    }


    /**
     * This factory method creates a new ClientModelConnector and
     * performs the ping to get the clock difference. Finally the
     * opcode listener thread is launched.
     *
     * @param hostname the hostname of the server
     * @param local tells whether the server is local or not: if local is
     * <code>true</code> clock difference will not be determined
     *
     * @throws IOException if the network fails to run the operations
     */
    public static ClientModelConnector 
        createClientModelConnector(String hostname, boolean local,
                                   String playerName) 
        throws IOException {

        ClientModelConnector cmc = 
            new ClientModelConnector(hostname, local, playerName);
        if (!local) {
            cmc.ping();
        }

        cmc.writePlayerName(playerName);
        cmc.flush();

        //new Thread(cmc, threadName).start();
        return cmc;
    }

    
    /**
     * JVM runs this method when start() is called for this thread.
     * Starts to listen for the opcodes.
     */
    public void run() {
        boolean doLoop = true;
        try {
            int opCode = 0;

            while(doLoop) {
                // at first we get the opCode, which is one byte long
                opCode = ois.read();
                
                switch(opCode) {

                case ServerModelConnector.OC_GAME_OVER:
                case ServerModelConnector.OC_SERVER_CLOSE:
                case ServerModelConnector.OC_CLIENT_CLOSE:
                    internalClose(opCode);
                    doLoop = false;
                    break;

                case ServerModelConnector.OC_ARENA:
                    readArena();
                    break;

                case ServerModelConnector.OC_INVALIDATE_BULLET:
                    readInvalidateBullet();
                    break;

                case ServerModelConnector.OC_INIT_BULLET:
                    readInitBullet();
                    break;

                case ServerModelConnector.OC_BLOW_UP_ASTEROID:
                    readBlowUpAsteroid();
                    break;

                case ServerModelConnector.OC_ASTEROID_LOCATION_AND_ANGLE:
                    readAsteroidLocationAndAngle();
                    break;

                case ServerModelConnector.OC_BLOW_UP_SHIP:
                    readBlowUpShip();
                    break;

                case ServerModelConnector.OC_SHIPS_LEFT:
                    readShipsLeft();
                    break;

                case ServerModelConnector.OC_SHIP_STATE:
                    readShipState();
                    break;

                case ServerModelConnector.OC_SCORE:
                    readScore();
                    break;

                case ServerModelConnector.OC_PLAYER_NAME:
                    readPlayerName();
                    break;
                    
                case ServerModelConnector.OC_EOF:
                    // end of network stream
                    throw new IOException("Network connection ended");

                default:
                    // invalid opCode
                    throw new IOException("Error in network connection");
                }
            }

        } catch(IOException ioe) {
            Runner.fatalError("Error in network connection!");
        }
    }    


    /**
     * Closes the the ship state pump and output stream, if they are not 
     * already closed. Finally closes also the input stream. If the close 
     * procedure is originated from this client, an <code>exit()</code> 
     * is also done. This method is called <b>only</b> from the opcode 
     * listener loop, when it receives a close opcode.
     *
     * @param opCode the opCode which initiated the close procedure 
     * 
     * @throws IOException if the network fails to run the operations
     */
    private synchronized void internalClose(int opCode) throws IOException {

        if (this.pump != null) {
            ShipStatePump tempPump = this.pump;
            this.pump = null;
            
            tempPump.stop();
        }

        if (this.oos != null) {
            ObjectOutputStream tempOos = this.oos;
            this.oos = null;

            tempOos.write(opCode);
            tempOos.flush();
            tempOos.close();
        }

        this.ois.close();
        this.s.close();


        // if it's a client originated close procedure from this client, 
        // do an exit()
        if (opCode == ServerModelConnector.OC_CLIENT_CLOSE && 
            this.shipIndex == CLIENT_SHIP_INDEX) {
            System.exit(0);

        } else if ( (opCode == ServerModelConnector.OC_CLIENT_CLOSE &&
                     this.shipIndex == SERVER_SHIP_INDEX) ||
                    (opCode == ServerModelConnector.OC_SERVER_CLOSE &&
                     this.shipIndex == CLIENT_SHIP_INDEX) ) {
            // other player quits or game's over
            this.otherPlayerHasQuit = true;

            if (this.spaceArea != null) {
                // inform the player about the close (if it's not a gameover)
                this.spaceArea.otherPlayerQuit();
            }
        } else if (opCode == ServerModelConnector.OC_GAME_OVER) {
            this.gameOver = true;
            if (this.spaceArea != null) {
                // tell the space area to draw the end credits
                this.spaceArea.gameOver();
            }
        }
    }

    /**
     * Closes the the ship state pump and output stream, if they are not 
     * already closed and calls the other connector to close its streams too. 
     * If this is the second close procedure, do an <code>exit()</code>.
     *
     * @param opCode the opCode which initiated the close procedure 
     * 
     * @throws IOException if the network fails to run the operations
     */
    private synchronized void close(int opCode) throws IOException {

        if (this.pump != null) {
            ShipStatePump tempPump = this.pump;
            this.pump = null;
            
            tempPump.stop();
        }

        if (this.oos != null) {
            ObjectOutputStream tempOos = this.oos;
            this.oos = null;
            
            tempOos.write(opCode);
            tempOos.flush();
            // output stream should not be closed, because input stream
            // does not like that (weird, isn't it?)
            //tempOos.close();
        }
        
        // if other player has quit earlier OR game is over, do an exit()
        if (this.otherPlayerHasQuit || this.gameOver) {
            System.exit(0);
        }
    }


    /**
     * A public method for initiating the close procedure from outside this
     * connector
     *
     * @throws IOException if the network fails to run the operations
     */
    public synchronized void close() throws IOException {
        if (this.shipIndex == SERVER_SHIP_INDEX) {
            this.close(ServerModelConnector.OC_SERVER_CLOSE);
        } else if (this.shipIndex == CLIENT_SHIP_INDEX) {
            this.close(ServerModelConnector.OC_CLIENT_CLOSE);
        }
    }


    /**
     * Saves the space area reference.
     * @param spaceArea the space area
     */
    public final void registerSpaceArea(SpaceArea spaceArea) {
        this.spaceArea = spaceArea;
    }


    /**
     * Sends the score board information to space area and marks a need
     * for update.
     */
    private final void updateScoreBoard() {
        String[] shipsLeftString = new String[2];

        // shipsLeft < 0, game is over and we lie to the space area about the
        // amount of ships left, because we don't want to show negative numbers
        for (int i = 0; i < 2; i++) {
            if (this.shipsLeft[i] < 0) {
                shipsLeftString[i] = "0";
            } else {
                shipsLeftString[i] = "" + this.shipsLeft[i];
            }
        }
            
        if(this.spaceArea != null) {
            this.spaceArea.updateScoreBoard(this.playerNames[0],
                                            this.playerNames[1],
                                            "" + this.scoreTable[0],
                                            "" + this.scoreTable[1],
                                            shipsLeftString[0],
                                            shipsLeftString[1]);
        }
    }


    /**
     * Flushes the network output stream
     *
     * @throws IOException if the network fails to run the operations
     */
    public synchronized void flush() throws IOException {
        if (this.oos != null) {
            this.oos.flush();
        }
    }


    /**
     * Getter method for the arena
     *
     * @return the connector's arena
     */
    public Arena getArena() {
        return this.arena;
    }


    /**
     * Getter method for the <code>shipIndex</code>
     *
     * @return the connector's <code>shipIndex</code>
     */
    public int getShipIndex() {
        return this.shipIndex;
    }


    /**
     * Getter method for the <code>playerName</code>
     *
     * @param shipIndex the player's <code>shipIndex</code>
     * @return the player name
     */
    public final String getPlayerName(int shipIndex) {
        return this.playerNames[shipIndex];
    }


    /**
     * Setter method for this client's <code>playerName</code>
     *
     * @param playerName the player name
     */
    public final void setPlayerName(String playerName) {
        this.playerNames[this.shipIndex] = playerName;
    }


    /**
     * Getter method for the <code>shipsLeft</code>
     *
     * @param shipIndex the player's <code>shipIndex</code>
     * @return the ships left for this player
     */
    public final int getShipsLeft(int shipIndex) {
        return this.shipsLeft[shipIndex];
    }


    /**
     * Getter method for the <code>scoreTable</code>
     *
     * @param shipIndex the player's <code>shipIndex</code>
     * @return the player's score
     */
    public final int getScore(int shipIndex) {
        return this.scoreTable[shipIndex];
    }


    /**
     * Getter method for the <code>inputHandler</code>
     *
     * @return the connector's <code>inputHandler</code>
     */
    public InputHandler getInputHandler() {
        return this.inputHandler;
    }


    /**
     * Returns the valid server time, which is the valid game time.
     *
     * @return the server time
     */
    public long currentServerMillis() {
        return System.currentTimeMillis() + this.clockDiff;
    }

    
    /**
     * Determines the difference between client and server machine clocks 
     * using socket connections and ObjectInput/OutputStreams.
     * You get the correct gameTime = serverTime in client by adding 
     * up the difference and clientTime: 
     * <code>gateTime = clientTime + difference.</code>
     * The ping() method should be called <b>only</b> from the 
     * createClientModelConnector() factory method.
     *
     * @throws IOException if the network fails to run the operations
     */
    private void ping() throws IOException {
        long avgClockDiff = 0;

        long [] clockDiff = new long[PING_COUNT];
        long [] transmissionDelay = new long[PING_COUNT];
        long begin = 0;
        long end = 0;

        for(int i = 0; i < PING_COUNT; i++) {

            begin = System.currentTimeMillis();
            oos.writeLong(System.currentTimeMillis());
            oos.flush();
                
            // difference = serverTime - clientTime
            clockDiff[i] = ois.readLong();
            transmissionDelay[i] = System.currentTimeMillis() - begin;
            try {
                Thread.sleep(PING_SLEEP_TIME);
            } catch(InterruptedException ie) {
            }
        }

        // add up
        long avgTransmissionDelay = 0;
        for(int i = 0; i < PING_COUNT; i++) {
            avgClockDiff += clockDiff[i];
            avgTransmissionDelay += transmissionDelay[i];
        }
            
        // count averages
        avgClockDiff /= PING_COUNT;
        avgTransmissionDelay /= (PING_COUNT * 2);

        if (avgTransmissionDelay > avgClockDiff) {
            avgClockDiff = 0;
        } else {
            if (avgClockDiff < 0) {
                avgClockDiff += avgTransmissionDelay;
            } else {
                avgClockDiff -= avgTransmissionDelay;
            }
        }
        
        this.clockDiff = avgClockDiff;
    }
    


    /**
     * Receives the Arena object to the client
     *
     * @throws IOException if the network fails to run the operations
     */
    private final void readArena() throws IOException {
        try {
            synchronized(this) {
                this.arena = (Arena)ois.readObject();
            }
        } catch(ClassNotFoundException cnfe) {
            Runner.fatalError("Error in Arena class!\n");
        }

        long timeToSleep = 
            this.arena.getLevelStarts() - this.currentServerMillis();
        if(timeToSleep > 0) {
            try {
                Thread.sleep(timeToSleep);
            } catch(InterruptedException ie) {
                // carry on
            }
        }
        this.spaceArea.registerArena(this.arena);

        SpaceShip ship = this.arena.getSpaceShips()[this.shipIndex];
        
        // save the reference of this ClientModelConnector to the local
        // space ship
        ship.registerConnector(this);
        this.inputHandler.registerShip(ship);
        this.pump.registerShip(ship);

	// refresh the score board, so that the player names 
        // will be drawn correctly
	this.updateScoreBoard();

        // explicit garbage collection
        System.gc();
    }


    /**
     * Marks a bullet non-flying (not visible).
     *
     * @throws IOException if the network fails to run the operations
     */
    private final void readInvalidateBullet() throws IOException {
        int shipIndex = ois.readInt();
        int bulletIndex = ois.readInt();

        this.arena.getSpaceShips()[shipIndex].
            getBullets()[bulletIndex].invalidate();
    }


    /**
     * Tells the client to blow up asteroid number <code>index</code>
     *
     * @throws IOException if the network fails to run the operations
     */
    private final void readBlowUpAsteroid() throws IOException {
        int index = ois.readInt();
        double locationX = ois.readDouble();
        double locationY = ois.readDouble();        
        long timestamp = ois.readLong();
        Asteroid.blowUpAsteroid(this.arena.getAsteroids(),
                                index,
                                locationX,
                                locationY,
                                timestamp,
                                false);
    }


    /**
     * Tells the client the location and angle of asteroid number 
     * <code>index</code>
     *
     * @throws IOException if the network fails to run the operations
     */
    private final void readAsteroidLocationAndAngle() throws IOException {
        int index = ois.readInt();
        double locationX = ois.readDouble();
        double locationY = ois.readDouble();
        double angle = ois.readDouble();
        long timestamp = ois.readLong();
        this.arena.getAsteroids()[index].setLocationAndAngle(locationX,
                                                             locationY,
                                                             angle,
                                                             timestamp);
    }
    
    /**
     * Tells the server that a bullet has been fired by the ship
     * controlled by this client.
     *
     * @param bulletIndex the index of the bullet
     * @param locationX the x coordinate at which the bullet was fired
     * @param locationY the y coordinate at which the bullet was fired
     * @param velocityX the x component of the velocity of the bullet
     * @param velocityY the y component of the velocity of the bullet
     * @param lineDeltaX the x difference of the between the center of
     *                   bullet the end of the bullet line
     * @param lineDeltaY the x difference of the between the center of
     *                   bullet the end of the bullet line
     * @param timestamp the moment at which the bullet was fired
     * @param expires the moment in time at which the bullet expires
     *
     * @throws IOException if the network fails to run the operations
     */
    public synchronized final void writeBulletFired(int bulletIndex,
                                                    double locationX,
                                                    double locationY,
                                                    double velocityX,
                                                    double velocityY,
                                                    double lineDeltaX,
                                                    double lineDeltaY,
                                                    long timestamp,
                                                    long expires) 
        throws IOException {

        if (this.oos != null) {
            oos.write(ServerModelConnector.OC_BULLET_FIRED);
            
            oos.writeInt(bulletIndex);
            oos.writeDouble(locationX);
            oos.writeDouble(locationY);
            oos.writeDouble(velocityX);
            oos.writeDouble(velocityY);
            oos.writeDouble(lineDeltaX);
            oos.writeDouble(lineDeltaY);
            
            oos.writeLong(timestamp);
            oos.writeLong(expires);
        }
    }


    /**
     * Tells the client that a bullet has been fired by the ship
     * controlled by the other client. 
     *
     * @throws IOException if the network fails to run the operations
     */
    private final void readInitBullet() throws IOException {
        int shipIndex = ois.readInt();
        int bulletIndex = ois.readInt();
        double locationX = ois.readDouble();
        double locationY = ois.readDouble();
        double velocityX = ois.readDouble();
        double velocityY = ois.readDouble();
        double lineDeltaX = ois.readDouble();
        double lineDeltaY = ois.readDouble();
        
        long timestamp = ois.readLong();
        long expires = ois.readLong();
        
        // initialize the bullet in the client model
        Bullet b;
        b = this.arena.getSpaceShips()[shipIndex].getBullets()[bulletIndex];
        b.initFromRemote(locationX,
                         locationY,
                         velocityX,
                         velocityY,
                         lineDeltaX,
                         lineDeltaY,
                         timestamp,
                         expires);
    }


    /**
     * Sends the state of the space ship to the server.
     *
     * @param locationX the x coordinate of the space ship
     * @param locationY the y coordinate of the space ship
     * @param velocityX the x component of the velocity of the space ship
     * @param velocityY the y component of the velocity of the space ship
     * @param angle the rotational angle of the space ship
     * @param timestamp the current server (game) time
     *
     * @throws IOException if the network fails to run the operations
     */
    public synchronized final void writeShipState(double locationX,
                                                  double locationY,
                                                  double velocityX,
                                                  double velocityY,
                                                  double angle,
                                                  long timestamp)
        throws IOException {
        
        if (this.oos != null) {
            oos.write(ServerModelConnector.OC_SHIP_STATE);
            
            oos.writeInt(this.shipIndex);
            oos.writeDouble(locationX);
            oos.writeDouble(locationY);
            oos.writeDouble(velocityX);
            oos.writeDouble(velocityY);
            oos.writeDouble(angle);
            
            oos.writeLong(timestamp);
        }
    }


    /**
     * Tells the client the state of the other space ship.
     *
     * @throws IOException if the network fails to run the operations
     */
    private final void readShipState() throws IOException {
        int shipIndex = ois.readInt();
        double locationX = ois.readDouble();
        double locationY = ois.readDouble();
        double velocityX = ois.readDouble();
        double velocityY = ois.readDouble();
        double angle = ois.readDouble();
        
        long timestamp = ois.readLong();
        
        this.arena.getSpaceShips()[shipIndex].setStateFromRemote(locationX,
                                                                 locationY,
                                                                 velocityX,
                                                                 velocityY,
                                                                 angle,
                                                                 timestamp);
    }


    /**
     * Tells the client that the ship at <code>shipIndex</code> has
     * blown up.
     *
     * @throws IOException if the network fails to run the operations
     */
    private final void readBlowUpShip() throws IOException {
        int shipIndex = ois.readInt();
        long newAppears = ois.readLong();
        this.arena.getSpaceShips()[shipIndex].explode(newAppears);
    }

    /**
     * Tells the client the number of ships a player has left.
     *
     * @throws IOException if the network fails to run the operations
     */    
    private final void readShipsLeft() throws IOException {
        int shipIndex = ois.readInt();
        int shipsLeft = ois.readInt();
        this.shipsLeft[shipIndex] = shipsLeft;
        this.updateScoreBoard();
    }

    /**
     * Read the new scores of a player.
     *
     * @throws IOException if the network fails to run the operations
     */
    private final void readScore() throws IOException {
        int shipIndex = ois.readInt();
        int score = ois.readInt();
        
        this.scoreTable[shipIndex] = score;
        this.updateScoreBoard();
    }


    /**
     * Read the player name and update it to the score board
     *
     * @throws IOException if the network fails to run the operations
     */
    private final void readPlayerName() throws IOException {
        int shipIndex = ois.readInt();

        try {
            this.playerNames[shipIndex] = (String) ois.readObject();
            this.updateScoreBoard();
        } catch(ClassNotFoundException cnfe) {
            Runner.fatalError("Error in Player name!");
        }
    }


    /**
     * Write this client's <code>playerName</code>
     *
     * @param playerName the player name string
     *
     * @throws IOException if the network fails to run the operations
     */
    private synchronized final void writePlayerName(String playerName) 
        throws IOException { 

        if (this.oos != null) {
            oos.writeObject(playerName);
        }
    }
}
