/*
 * Copyright 2002 Oskar Ojala, Yrj Kari-Koskinen and Henri Sivonen.
 *
 * The contents of this file are subject to the Artistic License.
 * You may not use this file except in compliance with the Artistic
 * License (which you should have received along with this file).
 */

package dualroids;

import java.awt.*;
import java.awt.event.*;

import javax.swing.*;
import javax.swing.text.*;

/**
 * This class shows the setup dialogs to the user and starts the game.
 * The <code>fatalError()</code> method is used for all errors
 * reported to the user.
 *
 * @author Oskar Ojala, omojala@cc.hut.fi
 * @author Yrj Kari-Koskinen, yrjo.kari-koskinen@hut.fi
 * @author Henri Sivonen, hsivonen@iki.fi
 */
public class Runner extends JFrame implements ActionListener {

    /** the server thread */
    private Server server;

    /** the client thread */
    private Client client;

    /** the field for player name in the main dialog */
    private JTextField playerNameField; 

    /** the buttons in the main dialog */
    private JButton connectButton, actButton;

    /** the host name string */
    private String host;

    /** the player name string */
    private String playerName;

    /** a static instance of the Runner (singleton pattern) */
    private static Runner theInstance;

    /** a constant indicating whether text should be antialiased */
    public static Object TEXT_ANTIALIASING = 
        RenderingHints.VALUE_TEXT_ANTIALIAS_ON;

    /**
     * the constructor, which should only be called from the 
     * <code>getInstance() method (singleton pattern).
     */
    private Runner() {
        this.server = null;
        this.client = null;
    }


    /**
     * This assures that no one can create more than one instance 
     * of Runner (singleton pattern)
     *
     * @return If theInstance is <code>null</code>, the Runner is created 
     * and returned. Otherwise the previously created instance is returned
     */
    public static Runner getInstance() {
        if (theInstance == null) {
            theInstance = new Runner();
        }
        return theInstance;
    }


    /**
     * Setup screen, which asks for the player's name and gives two choices:
     * <code>Connect to server</code> and <code>Act as server</code>.
     */
    private void setup() {
        setTitle("Dualroids Setup");
 
        setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
 
        playerNameField = new JTextField(15);
        playerNameField.setActionCommand("Player");
        
        JLabel playerLabel = new JLabel("Player name:");
        playerLabel.setLabelFor(playerNameField);

        connectButton = new JButton("Connect to server");
        actButton = new JButton("Act as server");
        connectButton.addActionListener(this);
        actButton.addActionListener(this);
        
        JPanel textControlsPane = new JPanel();
        textControlsPane.setLayout(new BoxLayout(textControlsPane,
                                                 BoxLayout.X_AXIS));
        textControlsPane.add(playerLabel);
        textControlsPane.add(Box.createRigidArea(new Dimension(8, 0)));
        textControlsPane.add(playerNameField);

        JPanel buttonsPane = new JPanel();
        buttonsPane.setLayout(new BoxLayout(buttonsPane,
                                            BoxLayout.X_AXIS));
        buttonsPane.add(connectButton);
        buttonsPane.add(Box.createRigidArea(new Dimension(12, 0)));
        buttonsPane.add(actButton);
        
        JPanel contentPane = new JPanel();
        contentPane.setBorder(BorderFactory.createEmptyBorder(20,20,20,20));
        contentPane.setLayout(new BoxLayout(contentPane, BoxLayout.Y_AXIS));
        contentPane.add(textControlsPane);
        contentPane.add(Box.createRigidArea(new Dimension(0, 18)));
        contentPane.add(buttonsPane);
 
        setContentPane(contentPane);
        pack();
        setResizable(false);
        setVisible(true);
        playerNameField.requestFocus();
    }

    /**
     * This takes care of the frame's buttons and invokes suitable methods.
     *
     * @param event the event
     */
    public void actionPerformed(ActionEvent event) {
        if (event.getSource() == connectButton) {
            connectToServer();
        }
        else if (event.getSource() == actButton) {
            actAsServer();
        }
    }


    /**
     * Gathers the necessary information to connect to the server and
     * connects to the server and starts the game.
     */
    private void connectToServer() {
        if (server != null) {
            JOptionPane.
                showMessageDialog(this,
                                  "Server is already running",
                                  "Error",
                                  JOptionPane.ERROR_MESSAGE);
            return;
        }
            
        if (this.client == null) {
            this.playerName = playerNameField.getText();
            
            if (this.playerName == null || this.playerName.equals("")) {
                this.playerName = "Player 2";
                playerNameField.setText(this.playerName);
            }
            this.host = 
                JOptionPane.showInputDialog(this, "Host name: ");
            
            // host=null, when cancel is pressed
            while (this.host != null && this.host.equals("")) {
                JOptionPane.
                    showMessageDialog(this,
                                      "Host name needed for connecting",
                                      "No host name",
                                      JOptionPane.ERROR_MESSAGE);
                this.host = JOptionPane.showInputDialog(this, "Host name: ");
            }
        }

        if (this.host != null) {

            // don't make duplicate clients
            if (this.client == null) {
                // client handles the connecting to the server
                this.client = new Client(this.host, false,
                                         this.playerName);
                new Thread(this.client, "Client").start();
            }

            Object[] possibleValues = { "Cancel" };
            // if cancel is pressed
            if(JOptionPane.showOptionDialog(this,
                                            "Connecting to server " + 
                                            this.host + "...",
                                            "Connect",
                                            JOptionPane.DEFAULT_OPTION,
                                            JOptionPane.INFORMATION_MESSAGE, 
                                            null,
                                            possibleValues, 
                                            possibleValues[0])
               == 0) {
                // then exit
                this.dispose();
                System.exit(-1);
            }
        }
    }

    /**
     * Starts server and waits for connections, then starts the game.
     */
    private void actAsServer() {
        if (this.client != null) {
            JOptionPane.
                showMessageDialog(this,
                                  "Client is already running",
                                  "Error",
                                  JOptionPane.ERROR_MESSAGE);
            return;
        }

        this.playerName = playerNameField.getText();

        if (this.playerName == null || this.playerName.equals("")) {
            this.playerName = "Player 1";
            playerNameField.setText(this.playerName);
        }

        // build a server, but not a second one
        if (this.server == null) {
            this.server = new Server(this.playerName);
            new Thread(this.server, "Server").start();
        }

        Object[] possibleValues = { "Cancel" };
        int button = 
            JOptionPane.showOptionDialog(this,
                                         "Waiting for 2nd player...",
                                         "Connect",
                                         JOptionPane.DEFAULT_OPTION,
                                         JOptionPane.INFORMATION_MESSAGE, 
                                         null,
                                         possibleValues, 
                                         possibleValues[0]);
        // if cancel is pressed
        if (button == 0) {
            // then exit
            this.dispose();
            System.exit(-1);
        } 
    }


    /**
     * This method should be used whenever a fatal error occurs and we want
     * to inform the user about it. This method hides the main game window
     * shows the error message dialog and finally does an <code>exit()</code>.
     * This method can be called from anywhere like this:
     * <code>Runner.fatalError("Something went wrong.");</code>
     *
     * @param text the text to be shown in the error message dialog
     */
    public static synchronized final void fatalError(String text) {
        GameFrame gf = GameFrame.getInstance();
        if (gf != null && gf.isShowing()) {
            gf.dispose();
        }

        JOptionPane.
            showMessageDialog(null, text, "Error!",
                              JOptionPane.ERROR_MESSAGE);

        System.exit(-1);
    }


    /**
     * The main method for starting the game. Command line parameters are
     * handled and the <code>Runner</code>'s <code>setup()</code> method 
     * is called.
     *
     * @param args the command line parameters: available parameters are:
     * <code>--no-db</code> which disables the double buffering and
     * <code>--no-antialiasing</code> which disables the antialiasing of 
     * texts in the game area
     */
    public static void main(String[] args) {
        // handle command line parameters
        for (int i = 0; i < args.length; i++) {
        
            // disable double buffering
            // http://java.sun.com/docs/books/performance/1st_edition/Additional_Material/Swing_and_Remote-X/swing_and_remote-x.html
            if (args[i].equals("--no-db")) {
                RepaintManager.currentManager(null).
                    setDoubleBufferingEnabled(false);
                System.out.println("Double buffering disabled");
            }

            // disable antialiasing
            if (args[i].equals("--no-antialias")) {
                TEXT_ANTIALIASING = RenderingHints.VALUE_TEXT_ANTIALIAS_OFF;
                System.out.println("Text antialiasing disabled");
            }
        }

        Runner runner = getInstance();
        runner.setup();
    }
}
