/*
 * Copyright 2002 Henri Sivonen.
 *
 * The contents of this file are subject to the Artistic License.
 * You may not use this file except in compliance with the Artistic
 * License (which you should have received along with this file).
 */

package dualroids;

import java.io.Serializable;
import java.awt.*;
import java.awt.geom.*;

/**
 * The common superclass for all spatial objects.
 *
 * @author Henri Sivonen, hsivonen@iki.fi
 */
public abstract class SpatialObject implements Serializable {

    /**
     * Moves <code>value</code> to the range
     * <code>0.0</code>...<code>Arena.WIDTH</code> so that the coordinate
     * wraps around.
     *
     * @param value the value to wrap
     * @return the wrapped value
     */
    public static final double moveToRangeX(double value) {
        while(value < 0.0) {
            value += Arena.WIDTH;
        }
        return value % Arena.WIDTH;
    }

    /**
     * Moves <code>value</code> to the range
     * <code>0.0</code>...<code>Arena.HEIGHT</code> so that the coordinate
     * wraps around.
     *
     * @param value the value to wrap
     * @return the wrapped value
     */
    public static final double moveToRangeY(double value) {
        while(value < 0.0) {
            value += Arena.HEIGHT;
        }
        return value % Arena.HEIGHT;
    }

    /**
     * The index of black in <code>COLOR_TABLE</code>.
     */
    public static final int COLOR_TABLE_BLACK = 0;

    /**
     * The index of white in <code>COLOR_TABLE</code>.
     */
    public static final int COLOR_TABLE_WHITE = 16;

    /**
     * The index of 50% gray in <code>COLOR_TABLE</code>.
     */
    public static final int COLOR_TABLE_GRAY_50 = 8;

    /**
     * 17 shades of gray ready to use in <code>Asteroid</code> and 
     * <code>SpaceShip</code> objects
     */
    public static final Color[] COLOR_TABLE = { Color.black,
                                               new Color(16, 16, 16),
                                               new Color(32, 32, 32),
                                               new Color(48, 48, 48),
                                               new Color(64, 64, 64),
                                               new Color(80, 80, 80),
                                               new Color(96, 96, 96),
                                               new Color(112, 112, 112),
                                               new Color(128, 128, 128),
                                               new Color(144, 144, 144),
                                               new Color(160, 160, 160),
                                               new Color(176, 176, 176),
                                               new Color(192, 192, 192),
                                               new Color(208, 208, 208),
                                               new Color(224, 224, 224),
                                               new Color(240, 240, 240),
                                               Color.white };
    
    /**
     * The moment in game (server) time for which the other fields are valid.
     */
    protected long timestamp;    
    
    /**
     * The untranslated geometric representation of the spatial object
     */
    protected transient GeneralPath geometry;

    /**
     * The x coordinate of the center of the spatial object at the
     * moment indicated by <code>timestamp</code>.
     */
    protected double locationX;

    /**
     * The y coordinate of the center of the spatial object at the
     * moment indicated by <code>timestamp</code>.
     */
    protected double locationY;

    /**
     * The x component of the valocity of the spatial object at the
     * moment indicated by <code>timestamp</code>.
     */
    protected double velocityX;

    /**
     * The y component of the valocity of the spatial object at the
     * moment indicated by <code>timestamp</code>.
     */
    protected double velocityY;

    /**
     * Returns a properly translated <code>Shape</code> representing
     * the spatial object at <code>time</code>
     *
     * @param time the moment in game (server) time for which the
     *             representation is requested
     * @return the <code>Shape</code> representation at <code>time</code>
     *         or null if the object isn't visible
     */
    public abstract Shape getTransformedShape(long time);

    /**
     * The <code>AffineTransform</code> used for calculating the translated
     * geometry. Stored here in order to avoid unnecessary feeding of the GC.
     */
    protected transient AffineTransform transform;
    
    /**
     * Contructs a new <code>SpatialObject</code>.
     *
     * @param locationX the x coordinate of the initial location
     * @param locationY the y coordinate of the initial location
     * @param velocityX the x component of the velocity
     * @param velocityY the y component of the velocity
     * @param angle the angle in radians
     * @param timestamp the moment in game (server) time for which the
     *                  other data is valid
     */
    protected SpatialObject(double locationX,
                            double locationY,
                            double velocityX,
                            double velocityY,
                            long timestamp)
    {
        this.locationX = moveToRangeX(locationX);
        this.locationY = moveToRangeY(locationY);
        this.velocityX = velocityX;
        this.velocityY = velocityY;
        this.timestamp = timestamp;
    }

    /**
     * Dummy constructor for keeping the serialization framework happy.
     */
    protected SpatialObject() { }
}
